"""Provides an object representing the current 'version' or 'release'
of Py as a whole.  Individual classes, such as the shell, filling and
interpreter, each have a revision property based on the CVS Revision."""

__author__ = "Patrick K. O'Brien <pobrien@orbtech.com>"

VERSION = '0.9.8'



# Name:         viewer.py
# Package:      wx.lib.pdfviewer
#
# Purpose:      A PDF report viewer class
#
# Author:       David Hughes     dfh@forestfield.co.uk
# Copyright:    Forestfield Software Ltd
# Licence:      Same as wxPython host

# History:      Created 17 Jun 2009
#
#               08 Oct 2011, Michael Hipp    michael@redmule.com
#               Added prompt, printer_name, orientation options to
#               pdfViewer.Print(). Added option to pdfViewer.LoadFile() to
#               accept a file-like object as well as a path string
#
# Tags:         phoenix-port, documented, unittest
#
#----------------------------------------------------------------------------

"""

This module provides the :class:`~wx.lib.pdfviewer.viewer.pdfViewer` to view PDF
files.
"""

import sys
import os
import time
import types
import copy
import shutil
from six import BytesIO, string_types

import wx

VERBOSE = True

try:
    # see http://pythonhosted.org/PyMuPDF - documentation & installation
    import fitz
    mupdf = True
    if VERBOSE: print('pdfviewer using PyMuPDF (GPL)')
except ImportError:
    mupdf = False
    try:
        # see http://pythonhosted.org/PyPDF2
        import PyPDF2
        from PyPDF2 import PdfFileReader
        from PyPDF2.pdf import ContentStream, PageObject
        from PyPDF2.filters import ASCII85Decode, FlateDecode
        if VERBOSE: print('pdfviewer using PyPDF2')
    except ImportError:
        msg = "PyMuPDF or PyPDF2 must be available to use pdfviewer"
        raise ImportError(msg)

GraphicsContext = wx.GraphicsContext
have_cairo = False
if not mupdf:
    try:
        import wx.lib.wxcairo as wxcairo
        import cairo
        from wx.lib.graphics import GraphicsContext
        have_cairo = True
        if VERBOSE: print('pdfviewer using Cairo')
    except ImportError:
        if VERBOSE: print('pdfviewer using wx.GraphicsContext')

    # New PageObject method added by Forestfield Software
    def extractOperators(self):
        """
        Locate and return all commands in the order they
        occur in the content stream
        """
        ops = []
        content = self["/Contents"].getObject()
        if not isinstance(content, ContentStream):
            content = ContentStream(content, self.pdf)
        for op in content.operations:
            if type(op[1] == bytes):
                op = (op[0], op[1].decode())
            ops.append(op)
        return ops
    # Inject this method into the PageObject class
    PageObject.extractOperators = extractOperators

    # If reportlab is installed, use its stringWidth metric. For justifying text,
    # where widths are cumulative, dc.GetTextExtent consistently underestimates,
    # possibly because it returns integer rather than float.
    try:
        from reportlab.pdfbase.pdfmetrics import stringWidth
        have_rlwidth = True
        if VERBOSE: print('pdfviewer using reportlab stringWidth function')
    except ImportError:
        have_rlwidth = False

#----------------------------------------------------------------------------

class pdfViewer(wx.ScrolledWindow):
    """
    View pdf file in a scrolled window.  Contents are read from PDF file
    and rendered in a GraphicsContext. Show visible window contents
    as quickly as possible then, when using pyPDF, read the whole file and build
    the set of drawing commands for each page. This can take time for a big file or if
    there are complex drawings eg. ReportLab's colour shading inside charts and a
    progress bar can be displayed by setting self.ShowLoadProgress = True (default)
    """
    def __init__(self, parent, nid, pos, size, style):
        """
        Default class constructor.

        :param wx.Window `parent`: parent window. Must not be ``None``;
        :param integer `nid`: window identifier. A value of -1 indicates a default value;
        :param `pos`: the control position. A value of (-1, -1) indicates a default position,
         chosen by either the windowing system or wxPython, depending on platform;
        :type `pos`: tuple or :class:`wx.Point`
        :param `size`: the control size. A value of (-1, -1) indicates a default size,
         chosen by either the windowing system or wxPython, depending on platform;
        :type `size`: tuple or :class:`wx.Size`
        :param integer `style`: the button style (unused);

        """
        wx.ScrolledWindow.__init__(self, parent, nid, pos, size,
                                style | wx.NO_FULL_REPAINT_ON_RESIZE)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)     # recommended in wxWidgets docs
        self.buttonpanel = None     # reference to panel is set by their common parent
        self._showLoadProgress = (not mupdf)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnResize)
        self.Bind(wx.EVT_SCROLLWIN, self.OnScroll)
        self.Bind(wx.EVT_IDLE, self.OnIdle)
        self.have_file = False
        self.resizing = False
        self.numpages = None
        self.zoomscale = -1     # fit page to screen width
        self.nom_page_gap = 20  # nominal inter-page gap (points)
        self.scrollrate = 20    # pixels per scrollbar increment
        self.page_buffer_valid = False
        self.page_after_zoom_change = None
        self.ClearBackground()

    def OnIdle(self, event):
        """
        Redraw on resize.
        """
        if self.resizing:
            self.Render()
            self.resizing = False
        event.Skip()

    def OnResize(self, event):
        """
        Buffer size change due to client area resize.
        """
        self.resizing = True
        event.Skip()

    def OnScroll(self, event):
        """
        Recalculate and redraw visible area. CallAfter is *essential*
        for coordination.
        """
        wx.CallAfter(self.Render)
        event.Skip()

    def OnPaint(self, event):
        """
        Refresh visible window with bitmap contents.
        """
        paintDC = wx.PaintDC(self)
        paintDC.Clear()         # in case buffer now smaller than visible window
        if hasattr(self, 'pdc'):
            paintDC.Blit(0, 0, self.winwidth, self.winheight, self.pdc,
                                                     self.xshift, self.yshift)

#----------------------------------------------------------------------------

    # This section defines the externally callable methods:
    # LoadFile, Save, Print, SetZoom, and GoPage
    # also the getter and setter for ShowLoadProgress
    # that is only applicable if using PyPDF2

    def LoadFile(self, pdf_file):
        """
        Read pdf file. Assume all pages are same size, for now.

        :param `pdf_file`: can be either a string holding
        a filename path or a file-like object.
        """
        def create_fileobject(filename):
            """
            Create and return a file object with the contents of filename,
            only used for testing.
            """
            f = open(filename, 'rb')
            stream = f.read()
            return BytesIO(stream)

        self.pdfpathname = ''
        if isinstance(pdf_file, string_types):
            # a filename/path string, save its name
            self.pdfpathname = pdf_file
            # remove comment from next line to test using a file-like object
            # pdf_file = create_fileobject(pdf_file)
        if mupdf:
            self.pdfdoc = mupdfProcessor(self, pdf_file)
        else:
            self.pdfdoc = pypdfProcessor(self, pdf_file, self.ShowLoadProgress)

        self.numpages = self.pdfdoc.numpages
        self.pagewidth = self.pdfdoc.pagewidth
        self.pageheight = self.pdfdoc.pageheight
        self.page_buffer_valid = False
        self.Scroll(0, 0)               # in case this is a re-LoadFile
        self.CalculateDimensions()      # to get initial visible page range
        # draw and display the minimal set of pages
        self.pdfdoc.DrawFile(self.frompage, self.topage)
        self.have_file = True
        # now draw full set of pages
        wx.CallAfter(self.pdfdoc.DrawFile, 0, self.numpages-1)

    def Save(self):
        "Save a copy of the pdf file if it was originally named"
        if self.pdfpathname:
            wild = "Portable document format (*.pdf)|*.pdf"
            dlg = wx.FileDialog(self, message="Save file as ...",
                                  wildcard=wild, style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if dlg.ShowModal() == wx.ID_OK:
                pathname = dlg.GetPath()
                shutil.copy(self.pdfpathname, pathname)
            dlg.Destroy()

    def Print(self, prompt=True, printer_name=None, orientation=None):
        """
        Print the pdf.

        :param boolean `prompt`: show the print dialog to the user (True/False). If
         False, the print dialog will not be shown and the pdf will be printed
         immediately. Default: True.
        :param string `printer_name`: the name of the printer that is to
         receive the printout. Default: as set by the O/S.
        :param `orientation`: select the orientation (:class:`wx.PORTRAIT` or
         :class:`wx.LANDSCAPE`) for the printout. Default: as set by the O/S.
        """
        pdd = wx.PrintDialogData()
        pdd.SetMinPage(1)
        pdd.SetFromPage(1)
        pdd.SetMaxPage(self.numpages)
        pdd.SetToPage(self.numpages)
        pdata = pdd.GetPrintData()
        if printer_name:
            pdata.SetPrinterName(printer_name)
        if orientation:
            pdata.SetOrientation(orientation)
        # PrintData does not return actual PrintQuality - it can't as printer_name not known
        # but it defaults to wx.PRINT_QUALITY_HIGH, overriding user's own setting for the
        # printer. However calling SetQuality with a value of 0 seems to leave the printer
        # setting untouched
        pdata.SetQuality(0)
        printer = wx.Printer(pdd)
        printout = pdfPrintout('', self)
        if (not printer.Print(self, printout, prompt=prompt) and
                       printer.GetLastError() == wx.PRINTER_ERROR):
            dlg = wx.MessageDialog(self, 'Unable to perform printing',
                              'Printer' , wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
        printout.Destroy()

    def SetZoom(self, zoomscale):
        """
        Positive integer or floating zoom scale will render the file at corresponding
        size where 1.0 is "actual" point size (1/72").
        -1 fits page width and -2 fits page height into client area
        Redisplay the current page(s) at the new size

        :param `zoomscale`: an integer or float

        """
        pagenow = self.frompage
        self.zoomscale = zoomscale
        self.page_buffer_valid = False
        # calling GoPage now will trigger rendering at the new size but the page location
        # will be calculated based on the old zoom scale - so save the required page number
        # and call GoPage again *after* rendering at the new size
        self.page_after_zoom_change = pagenow
        self.GoPage(pagenow)

    def GoPage(self, pagenum):
        """
        Go to page

        :param integer `pagenum`: go to the provided page number if it is valid

        """
        if pagenum > 0 and pagenum <= self.numpages:
            self.Scroll(0, pagenum*self.Ypagepixels/self.GetScrollPixelsPerUnit()[1] + 1)
        else:
            self.Scroll(0, 0)
        # calling Scroll sometimes doesn't raise wx.EVT_SCROLLWIN eg Windows 8 64 bit - so
        wx.CallAfter(self.Render)

    @property
    def ShowLoadProgress(self):
        """Property to control if file reading progress is shown (PyPDF2 only)"""
        return self._showLoadProgress

    @ShowLoadProgress.setter
    def ShowLoadProgress(self, flag):
        """Setter for showLoadProgress."""
        self._showLoadProgress = flag

#----------------------------------------------------------------------------

    # This section is concerned with rendering a sub-set of drawing commands on demand

    def CalculateDimensions(self):
        """
        Compute the required buffer sizes to hold the viewed rectangle and
        the range of pages visible. Set self.page_buffer_valid = False if
        the current set of rendered pages changes
        """
        self.frompage = 0
        self.topage = 0
        device_scale = wx.ClientDC(self).GetPPI()[0]/72.0   # pixels per inch/points per inch
        self.font_scale_metrics =  1.0
        self.font_scale_size = 1.0
        # for Windows only with wx.GraphicsContext the rendered font size is too big
        # in the ratio of screen pixels per inch to points per inch
        # and font metrics are too big in the same ratio for both for Cairo and wx.GC
        if wx.PlatformInfo[1] == 'wxMSW':
            self.font_scale_metrics = 1.0 / device_scale
            if not have_cairo:
                self.font_scale_size = 1.0 / device_scale

        self.winwidth, self.winheight = self.GetClientSize()
        if self.winheight < 100:
            return
        self.Ypage = self.pageheight + self.nom_page_gap
        if self.zoomscale > 0.0:
            self.scale = self.zoomscale * device_scale
        else:
            if int(self.zoomscale) == -1:   # fit width
                self.scale = self.winwidth / self.pagewidth
            else:                           # fit page
                self.scale = self.winheight / self.pageheight
        self.Xpagepixels = int(round(self.pagewidth*self.scale))
        self.Ypagepixels = int(round(self.Ypage*self.scale))

        # adjust inter-page gap so Ypagepixels is a whole number of scroll increments
        # and page numbers change precisely on a scroll click
        idiv = self.Ypagepixels/self.scrollrate
        nlo = idiv * self.scrollrate
        nhi = (idiv + 1) * self.scrollrate
        if nhi - self.Ypagepixels < self.Ypagepixels - nlo:
            self.Ypagepixels = nhi
        else:
            self.Ypagepixels = nlo
        self.page_gap = self.Ypagepixels/self.scale - self.pageheight

        self.maxwidth = max(self.winwidth, self.Xpagepixels)
        self.maxheight = max(self.winheight, self.numpages*self.Ypagepixels)
        self.SetVirtualSize((self.maxwidth, self.maxheight))
        self.SetScrollRate(self.scrollrate, self.scrollrate)

        xv, yv = self.GetViewStart()
        dx, dy = self.GetScrollPixelsPerUnit()
        self.x0, self.y0   = (xv * dx, yv * dy)
        self.frompage = int(min(self.y0/self.Ypagepixels, self.numpages-1))
        self.topage = int(min((self.y0+self.winheight-1)/self.Ypagepixels, self.numpages-1))
        self.pagebufferwidth = max(self.Xpagepixels, self.winwidth)
        self.pagebufferheight = (self.topage - self.frompage + 1) * self.Ypagepixels

        # Inform buttonpanel controls of any changes
        if self.buttonpanel:
            self.buttonpanel.Update(self.frompage, self.numpages,
                                      self.scale/device_scale)

        self.page_y0 = self.frompage * self.Ypagepixels
        self.page_x0 = 0
        self.xshift = self.x0 - self.page_x0
        self.yshift = self.y0 - self.page_y0
        if not self.page_buffer_valid:  # via external setting
            self.cur_frompage = self.frompage
            self.cur_topage = self.topage
        else:   # page range unchanged? whole visible area will always be inside page buffer
            if self.frompage != self.cur_frompage or self.topage != self.cur_topage:
                self.page_buffer_valid = False    # due to page buffer change
                self.cur_frompage = self.frompage
                self.cur_topage = self.topage
        return

    def Render(self):
        """
        Recalculate dimensions as client area may have been scrolled or resized.
        The smallest unit of rendering that can be done is the pdf page. So render
        the drawing commands for the pages in the visible rectangle into a buffer
        big enough to hold this set of pages. Force re-creating the page buffer
        only when client view moves outside it.
        With PyPDF2, use gc.Translate to render each page wrt the pdf origin,
        which is at the bottom left corner of the page.
        """
        if not self.have_file:
            return
        self.CalculateDimensions()
        if not self.page_buffer_valid:
            # Initialize the buffer bitmap.
            self.pagebuffer = wx.Bitmap(self.pagebufferwidth, self.pagebufferheight)
            self.pdc = wx.MemoryDC(self.pagebuffer)     # must persist

            gc = GraphicsContext.Create(self.pdc)       # Cairo/wx.GraphicsContext API

            # white background
            path = gc.CreatePath()
            path.AddRectangle(0, 0,
                                self.pagebuffer.GetWidth(), self.pagebuffer.GetHeight())
            gc.SetBrush(wx.WHITE_BRUSH)
            gc.FillPath(path)

            for pageno in range(self.frompage, self.topage+1):
                self.xpageoffset = 0 - self.x0
                self.ypageoffset = pageno*self.Ypagepixels - self.page_y0
                gc.PushState()
                if mupdf:
                    gc.Translate(self.xpageoffset, self.ypageoffset)
                    # scaling is done inside RenderPage
                else:

                    gc.Translate(self.xpageoffset, self.ypageoffset +
                                    self.pageheight*self.scale)
                    gc.Scale(self.scale, self.scale)
                self.pdfdoc.RenderPage(gc, pageno, scale=self.scale)
                # Show inter-page gap
                gc.SetBrush(wx.Brush(wx.Colour(180, 180, 180)))        #mid grey
                gc.SetPen(wx.TRANSPARENT_PEN)
                if mupdf:
                    gc.DrawRectangle(0, self.pageheight*self.scale,
                                 self.pagewidth*self.scale, self.page_gap*self.scale)
                else:
                    gc.DrawRectangle(0, 0, self.pagewidth, self.page_gap)
                gc.PopState()
            gc.PushState()
            gc.Translate(0-self.x0, 0-self.page_y0)
            self.RenderPageBoundaries(gc)
            gc.PopState()

        self.page_buffer_valid = True
        self.Refresh(0) # Blit appropriate area of new or existing page buffer to screen

        # ensure we stay on the same page after zoom scale is changed
        if self.page_after_zoom_change:
            self.GoPage(self.page_after_zoom_change)
            self.page_after_zoom_change = None

    def RenderPageBoundaries(self, gc):
        """
        Show non-page areas in grey.
        """
        gc.SetBrush(wx.Brush(wx.Colour(180, 180, 180)))        #mid grey
        gc.SetPen(wx.TRANSPARENT_PEN)
        gc.Scale(1.0, 1.0)
        extrawidth = self.winwidth - self.Xpagepixels
        if extrawidth > 0:
            gc.DrawRectangle(self.winwidth-extrawidth, 0, extrawidth, self.maxheight)
        extraheight = self.winheight - (self.numpages*self.Ypagepixels - self.y0)
        if extraheight > 0:
            gc.DrawRectangle(0, self.winheight-extraheight, self.maxwidth, extraheight)

#============================================================================

class mupdfProcessor(object):
    """
    Create an instance of this class to open a PDF file, process the contents of
    each page and render each one on demand using the GPL mupdf library, which is
    accessed via the python-fitz package bindings (version 1.9.1 or later)
    """
    def __init__(self, parent, pdf_file):
        """
        :param `pdf_file`: a File object or an object that supports the standard
        read and seek methods similar to a File object.
        Could also be a string representing a path to a PDF file.
        """
        self.parent = parent
        if isinstance(pdf_file, string_types):
            # a filename/path string, pass the name to fitz.open
            pathname = pdf_file
            self.pdfdoc = fitz.open(pathname)
        else:
            # assume it is a file-like object, pass the stream content to fitz.open
            # and a '.pdf' extension in pathname to identify the stream type
            pathname = 'fileobject.pdf'
            if pdf_file.tell() > 0:     # not positioned at start
                pdf_file.seek(0)
            stream = bytearray(pdf_file.read())
            self.pdfdoc = fitz.open(pathname, stream)

        self.numpages = self.pdfdoc.pageCount
        page = self.pdfdoc.loadPage(0)
        self.pagewidth = page.bound().width
        self.pageheight = page.bound().height
        self.page_rect = page.bound()
        self.zoom_error = False     #set if memory errors during render

    def DrawFile(self, frompage, topage):
        """
        This is a no-op for mupdf. Each page is scaled and drawn on
        demand during RenderPage directly via a call to page.getPixmap()
        """
        self.parent.GoPage(frompage)

    def RenderPage(self, gc, pageno, scale=1.0):
        " Render the set of pagedrawings into gc for specified page "
        page = self.pdfdoc.loadPage(pageno)
        matrix = fitz.Matrix(scale, scale)
        try:
            pix = page.getPixmap(matrix=matrix)   # MUST be keyword arg(s)
            bmp = wx.Bitmap.FromBufferRGBA(pix.width, pix.height, pix.samples)
            gc.DrawBitmap(bmp, 0, 0, pix.width, pix.height)
            self.zoom_error = False
        except (RuntimeError, MemoryError):
            if not self.zoom_error:     # report once only
                self.zoom_error = True
                dlg = wx.MessageDialog(self.parent, 'Out of memory. Zoom level too high?',
                              'pdf viewer' , wx.OK |wx.ICON_EXCLAMATION)
                dlg.ShowModal()
                dlg.Destroy()

#============================================================================

class pypdfProcessor(object):
    """
    Create an instance of this class to open a PDF file, process the contents of
    every page using PyPDF2 then render each one on demand
    """
    def __init__(self, parent, fileobj, showloadprogress):
        self.parent = parent
        self.showloadprogress = showloadprogress
        self.pdfdoc = PdfFileReader(fileobj)
        self.numpages = self.pdfdoc.getNumPages()
        page1 = self.pdfdoc.getPage(0)
        self.pagewidth = float(page1.mediaBox.getUpperRight_x())
        self.pageheight = float(page1.mediaBox.getUpperRight_y())
        self.pagedrawings = {}
        self.unimplemented = {}
        self.formdrawings = {}
        self.page = None
        self.gstate = None
        self.saved_state = None
        self.knownfont = False
        self.progbar = None

    # These methods interpret the PDF contents as a set of drawing commands

    def Progress(self, ptype, value):
        " This function is called at regular intervals during Drawfile"
        if ptype == 'start':
            pmsg = 'Reading pdf file'
            self.progbar = wx.ProgressDialog('Load file', pmsg, value, None,
                         wx.PD_AUTO_HIDE|
                            wx.PD_ESTIMATED_TIME|wx.PD_REMAINING_TIME)
        elif ptype == 'progress':
            self.progbar.Update(value)
        elif ptype == 'end':
            self.progbar.Destroy()

    def DrawFile(self, frompage, topage):
        """
        Build set of drawing commands from PDF contents. Ideally these could be drawn
        straight into a PseudoDC and the visible section painted directly into
        scrolled window, but we need to be able to zoom and scale the output quickly
        without having to rebuild the drawing commands (slow). So build our
        own command lists, one per page, into self.pagedrawings.
        """
        numpages_generated = 0
        rp = (self.showloadprogress and frompage == 0 and topage == self.numpages-1)
        if rp: self.Progress('start', self.numpages)
        for pageno in range(frompage, topage+1):
            self.gstate = pdfState()    # state is reset with every new page
            self.saved_state = []
            self.page = self.pdfdoc.getPage(pageno)
            numpages_generated += 1
            pdf_fonts = self.FetchFonts(self.page)
            self.pagedrawings[pageno] = self.ProcessOperators(
                                    self.page.extractOperators(), pdf_fonts)
            if rp: self.Progress('progress', numpages_generated)

        if rp: self.Progress('end', None)
        self.parent.GoPage(frompage)

    def RenderPage(self, gc, pageno, scale=None):
        """
        Render the set of pagedrawings
        In a pdf file, bitmaps are treated as being of unit width and height and
        are scaled via a previous ConcatTransform containing the corresponding width
        and height as scale factors. wx.GraphicsContext/Cairo appear not to respond to
        this so scaling is removed from transform and width & height are added
        to the Drawbitmap call.
        """
        drawdict = {'ConcatTransform': gc.ConcatTransform,
                    'PushState': gc.PushState,
                    'PopState': gc.PopState,
                    'SetFont': gc.SetFont,
                    'SetPen': gc.SetPen,
                    'SetBrush': gc.SetBrush,
                    'DrawText': gc.DrawText,
                    'DrawBitmap': gc.DrawBitmap,
                    'CreatePath': gc.CreatePath,
                    'DrawPath': gc.DrawPath }
        for drawcmd, args, kwargs in self.pagedrawings[pageno]:
            # scale font if requested by printer DC
            if drawcmd == 'SetFont' and hasattr(gc, 'font_scale'):
                args[0].Scale(gc.font_scale)
            if drawcmd == 'ConcatTransform':
                cm = gc.CreateMatrix(*args, **kwargs)
                args = (cm,)
            if drawcmd == 'CreatePath':
                gp = drawdict[drawcmd](*args, **kwargs)
                continue
            elif drawcmd == 'DrawPath':
                args = (gp, args[1])
            if drawcmd in drawdict:
                drawdict[drawcmd](*args, **kwargs)
                # reset font scaling in case RenderPage call is repeated
                if drawcmd == 'SetFont' and hasattr(gc, 'font_scale'):
                    args[0].Scale(1.0/gc.font_scale)
            else:
                pathdict = {'MoveToPoint': gp.MoveToPoint,
                            'AddLineToPoint': gp.AddLineToPoint,
                            'AddCurveToPoint': gp.AddCurveToPoint,
                            'AddRectangle': gp.AddRectangle,
                            'CloseSubpath': gp.CloseSubpath }
                if drawcmd in pathdict:
                    pathdict[drawcmd](*args, **kwargs)

    def FetchFonts(self, currentobject):
        " Return the standard fonts in current page or form"
        pdf_fonts = {}
        try:
            fonts = currentobject["/Resources"].getObject()['/Font']
            for key in fonts:
                pdf_fonts[key] = fonts[key]['/BaseFont'][1:]     # remove the leading '/'
        except KeyError:
            pass
        return pdf_fonts

    def ProcessOperators(self, opslist, pdf_fonts):
        """
        Interpret each operation in opslist and return in drawlist.
        """
        drawlist = []
        path = []
        for operand, operator in opslist :
            g = self.gstate
            if operator == 'cm' and operand:        # new transformation matrix
                # some operands need inverting because directions of y axis
                # in pdf and graphics context are opposite
                a, b, c, d, e, f = [float(n) for n in operand]
                drawlist.append(['ConcatTransform', (a, -b, -c, d, e, -f), {}])
            elif operator == 'q':       # save state
                self.saved_state.append(copy.deepcopy(g))
                drawlist.append(['PushState', (), {}])
            elif operator == 'Q':       # restore state
                self.gstate = self.saved_state.pop()
                drawlist.append(['PopState', (), {}])
            elif operator == 'RG':      # Stroke RGB
                rs, gs, bs = [int(float(n)*255) for n in operand]
                g.strokeRGB = wx.Colour(rs, gs, bs)
            elif operator == 'rg':      # Fill RGB
                rf, gf, bf = [int(float(n)*255) for n in operand]
                g.fillRGB = wx.Colour(rf, gf, bf)
            elif operator == 'K':       # Stroke CMYK
                rs, gs, bs = self.ConvertCMYK(operand)
                g.strokeRGB = wx.Colour(rs, gs, bs)
            elif operator == 'k':       # Fill CMYK
                rf, gf, bf = self.ConvertCMYK(operand)
                g.fillRGB = wx.Colour(rf, gf, bf)
            elif operator == 'w':       # Line width
                g.lineWidth = max(float(operand[0]), 1.0)
            elif operator == 'J':       # Line cap
                ix = float(operand[0])
                g.lineCapStyle = {0: wx.CAP_BUTT, 1: wx.CAP_ROUND,
                                              2: wx.CAP_PROJECTING}[ix]
            elif operator == 'j':       # Line join
                ix = float(operand[0])
                g.lineJoinStyle = {0: wx.JOIN_MITER, 1: wx.JOIN_ROUND,
                                              2: wx.JOIN_BEVEL}[ix]
            elif operator == 'd':       # Line dash pattern
                g.lineDashArray = [int(n) for n in operand[0]]
                g.lineDashPhase = int(operand[1])
            elif operator in ('m', 'c', 'l', 're', 'v', 'y', 'h'):    # path defining ops
                path.append([[float(n) for n in operand], operator])
            elif operator in ('b', 'B', 'b*', 'B*', 'f', 'F', 'f*',
                                           's', 'S', 'n'):    # path drawing ops
                drawlist.extend(self.DrawPath(path, operator))
                path = []
            elif operator == 'BT':      # begin text object
                g.textMatrix = [1, 0, 0, 1, 0, 0]
                g.textLineMatrix = [1, 0, 0, 1, 0, 0]
            elif operator == 'ET':      # end text object
                continue
            elif operator == 'Tm':      # text matrix
                g.textMatrix = [float(n) for n in operand]
                g.textLineMatrix = [float(n) for n in operand]
            elif operator == 'TL':      # text leading
                g.leading = float(operand[0])
            #elif operator == 'Tc':     # character spacing
            #    g.charSpacing = float(operand[0])
            elif operator == 'Tw':      # word spacing
                g.wordSpacing = float(operand[0])
            elif operator == 'Ts':      # super/subscript
                g.textRise = float(operand[0])
            elif operator == 'Td':      # next line via offsets
                g.textLineMatrix[4] += float(operand[0])
                g.textLineMatrix[5] += float(operand[1])
                g.textMatrix = copy.copy(g.textLineMatrix)
            elif operator == 'T*':      # next line via leading
                g.textLineMatrix[4] += 0
                g.textLineMatrix[5] -= g.leading if g.leading is not None else 0
                g.textMatrix = copy.copy(g.textLineMatrix)
            elif operator == 'Tf':      # text font
                g.font = pdf_fonts[operand[0]]
                g.fontSize = float(operand[1])
            elif operator == 'Tj':      # show text
                drawlist.extend(self.DrawTextString(
                                       operand[0].original_bytes.decode('latin-1')))
            elif operator == 'Do':      # invoke named XObject
                dlist = self.InsertXObject(operand[0])
                if dlist:               # may be unimplemented decode
                    drawlist.extend(dlist)
            elif operator == 'INLINE IMAGE':    # special pyPdf case + operand is a dict
                dlist = self.InlineImage(operand)
                if dlist:               # may be unimplemented decode
                    drawlist.extend(dlist)
            else:                       # report once
                if operator not in self.unimplemented:
                    if VERBOSE: print('PDF operator %s is not implemented' % operator)
                    self.unimplemented[operator] = 1

        # Fix bitmap transform. Move the scaling from any transform matrix that precedes
        # a DrawBitmap operation into the op itself - the width and height extracted from
        # the bitmap is the size of the original PDF image not the size it is to be drawn
        for k in range(len(drawlist)-1):
            if drawlist[k][0] == 'ConcatTransform' and drawlist[k+1][0] == 'DrawBitmap':
                ctargs = list(drawlist[k][1])
                bmargs = list(drawlist[k+1][1])
                bmargs[2] = -ctargs[3]          # y position
                bmargs[3] = ctargs[0]           # width
                bmargs[4] = ctargs[3]           # height
                ctargs[0] = 1.0
                ctargs[3] = 1.0
                drawlist[k][1] = tuple(ctargs)
                drawlist[k+1][1] = tuple(bmargs)
        return drawlist

    def SetFont(self, pdfont, size):
        """
        Returns :class:`wx.Font` instance from supplied pdf font information.
        """
        self.knownfont = True
        pdfont = pdfont.lower()
        if pdfont.count('courier'):
            family = wx.FONTFAMILY_MODERN
            font = 'Courier New'
        elif pdfont.count('helvetica'):
            family = wx.FONTFAMILY_SWISS
            font = 'Arial'
        elif pdfont.count('times'):
            family = wx.FONTFAMILY_ROMAN
            font = 'Times New Roman'
        elif pdfont.count('symbol'):
            family = wx.FONTFAMILY_DEFAULT
            font = 'Symbol'
        elif pdfont.count('zapfdingbats'):
            family = wx.FONTFAMILY_DEFAULT
            font = 'Wingdings'
        else:
            if VERBOSE: print('Unknown font %s' % pdfont)
            self.knownfont = False
            family = wx.FONTFAMILY_SWISS
            font = 'Arial'

        weight = wx.FONTWEIGHT_NORMAL
        if pdfont.count('bold'):
            weight = wx.FONTWEIGHT_BOLD
        style = wx.FONTSTYLE_NORMAL
        if pdfont.count('oblique') or pdfont.count('italic'):
            style = wx.FONTSTYLE_ITALIC
        return wx.Font(max(1, size), family, style, weight, faceName=font)

    def DrawTextString(self, text):
        """
        Draw a text string. Word spacing only works for horizontal text.

        :param string `text`: the text to draw

        """
        dlist = []
        g = self.gstate
        f0  = self.SetFont(g.font, g.fontSize)
        f0.Scale(self.parent.font_scale_metrics)
        f1  = self.SetFont(g.font, g.fontSize)
        f1.Scale(self.parent.font_scale_size)
        dlist.append(['SetFont', (f1, g.fillRGB), {}])
        if g.wordSpacing > 0:
            textlist = text.split(' ')
        else:
            textlist = [text,]
        for item in textlist:
            dlist.append(self.DrawTextItem(item, f0))
        return dlist

    def DrawTextItem(self, textitem, f):
        """
        Draw a text item.

        :param `textitem`: the item to draw
        :param `f`: the font to use for text extent measuring

        """
        dc = wx.ClientDC(self.parent)      # dummy dc for text extents
        g = self.gstate
        x = g.textMatrix[4]
        y = g.textMatrix[5] + g.textRise
        if g.wordSpacing > 0:
            textitem += ' '
        wid, ht, descend, x_lead = dc.GetFullTextExtent(textitem, f)
        if have_rlwidth and self.knownfont:   # use ReportLab stringWidth if available
            width = stringWidth(textitem, g.font, g.fontSize)
        else:
            width = wid
        g.textMatrix[4] += (width + g.wordSpacing)  # update current x position
        return ['DrawText', (textitem, x, -y-(ht-descend)), {}]

    def DrawPath(self, path, action):
        """
        Stroke and/or fill the defined path depending on operator.
        """
        dlist = []
        g = self.gstate
        acts = {'S':  (1, 0, 0),
                's':  (1, 0, 0),
                'f':  (0, 1, wx.WINDING_RULE),
                'F':  (0, 1, wx.WINDING_RULE),
                'f*': (0, 1, wx.ODDEVEN_RULE),
                'B':  (1, 1, wx.WINDING_RULE),
                'B*': (1, 1, wx.ODDEVEN_RULE),
                'b':  (1, 1, wx.WINDING_RULE),
                'b*': (1, 1, wx.ODDEVEN_RULE),
                'n':  (0, 0, 0) }
        stroke, fill, rule = acts[action]
        if action in ('s', 'b', 'b*'):
            path.append([[], 'h'])      # close path

        if stroke:
            if g.lineDashArray:
                style = wx.PENSTYLE_USER_DASH
            else:
                style = wx.PENSTYLE_SOLID
            cpen = wx.Pen(g.strokeRGB, g.lineWidth, style)
            cpen.SetCap(g.lineCapStyle)
            cpen.SetJoin(g.lineJoinStyle)
            if g.lineDashArray:
                cpen.SetDashes(g.lineDashArray)
            dlist.append(['SetPen', (cpen,), {}])
        else:
            dlist.append(['SetPen', (wx.TRANSPARENT_PEN,), {}])

        if fill:
            dlist.append(['SetBrush', (wx.Brush(g.fillRGB),), {}])
        else:
            dlist.append(['SetBrush', (wx.TRANSPARENT_BRUSH,), {}])

        dlist.append(['CreatePath', (), {}])
        for xylist, op in path:
            if op == 'm':           # move (to) current point
                x0 = xc = xylist[0]
                y0 = yc = -xylist[1]
                dlist.append(['MoveToPoint', (x0, y0), {}])
            elif op == 'l':         # draw line
                x2 = xylist[0]
                y2 = -xylist[1]
                dlist.append(['AddLineToPoint', (x2, y2), {}])
                xc = x2
                yc = y2
            elif op == 're':        # draw rectangle
                x = xylist[0]
                y = -xylist[1]
                w = xylist[2]
                h = xylist[3]
                retuple = (x, y-h, w, h)
                if h < 0.0:
                    retuple = (x, y, w, -h)
                dlist.append(['AddRectangle', retuple, {}])
            elif op in ('c', 'v', 'y'):         # draw Bezier curve
                args = []
                if op == 'v':
                    args.extend([xc, yc])
                args.extend([xylist[0], -xylist[1],
                                xylist[2], -xylist[3]])
                if op == 'y':
                    args.extend([xylist[2], -xylist[3]])
                if op == 'c':
                    args.extend([xylist[4], -xylist[5]])
                dlist.append(['AddCurveToPoint', args, {}])
            elif op == 'h':
                dlist.append(['CloseSubpath', (), {}])
        dlist.append(['DrawPath', ('GraphicsPath', rule), {}])
        return dlist

    def InsertXObject(self, name):
        """
        XObject can be an image or a 'form' (an arbitrary PDF sequence).
        """
        dlist = []
        xobject = self.page["/Resources"].getObject()['/XObject']
        stream = xobject[name]
        if stream.get('/Subtype') == '/Form':
            # insert contents into current page drawing
            if not name in self.formdrawings:       # extract if not already done
                pdf_fonts = self.FetchFonts(stream)
                x_bbox = stream.get('/BBox')
                matrix = stream.get('/Matrix')
                form_ops = ContentStream(stream, self.pdfdoc).operations
                oplist = [([], 'q'), (matrix, 'cm')]    # push state & apply matrix
                oplist.extend(form_ops)                 # add form contents
                oplist.append(([], 'Q'))                # restore original state
                self.formdrawings[name] = self.ProcessOperators(oplist, pdf_fonts)
            dlist.extend(self.formdrawings[name])
        elif stream.get('/Subtype') == '/Image':
            width = stream['/Width']
            height = stream['/Height']
            x_depth = stream['/BitsPerComponent']
            filters = stream["/Filter"]
            item = self.AddBitmap(stream._data, width, height, filters)
            if item:            # may be unimplemented
                dlist.append(item)
        return dlist

    def InlineImage(self, operand):
        """ operand contains an image"""
        dlist = []
        data = operand.get('data')
        settings = operand.get('settings')
        width = settings['/W']
        height = settings['/H']
        x_depth = settings['/BPC']
        filters = settings['/F']
        item = self.AddBitmap(data, width, height, filters)
        if item:            # may be unimplemented
            dlist.append(item)
        return dlist

    def AddBitmap(self, data, width, height, filters):
        """
        Add wx.Bitmap from data, processed by filters.
        """
        if '/A85' in filters or '/ASCII85Decode' in filters:
            data = ASCII85Decode.decode(data)
        if '/Fl' in filters or '/FlateDecode' in filters:
            data = FlateDecode.decode(data, None)
        if '/CCF' in filters or  '/CCITTFaxDecode' in filters:
            if VERBOSE:
                print('PDF operation /CCITTFaxDecode is not implemented')
            return []
        if '/DCT' in filters or '/DCTDecode' in filters:
            stream = BytesIO(data)
            image = wx.Image(stream, wx.BITMAP_TYPE_JPEG)
            bitmap = wx.Bitmap(image)
        else:
            try:
                bitmap = wx.Bitmap.FromBuffer(width, height, data)
            except:
                return []       # any error
        return ['DrawBitmap', (bitmap, 0, 0-height, width, height), {}]

    def ConvertCMYK(self, operand):
        """
        Convert CMYK values (0 to 1.0) in operand to nearest RGB.
        """
        c, m, y, k = operand
        r = round((1-c)*(1-k)*255)
        b = round((1-y)*(1-k)*255)
        g = round((1-m)*(1-k)*255)
        return (r, g, b)

#----------------------------------------------------------------------------

class pdfState(object):
    """
    Instance holds the current pdf graphics and text state. It can be
    saved (pushed) and restored (popped) by the owning parent
    """
    def __init__ (self):
        """
        Creates an instance with default values. Individual attributes
        are modified directly not via getters and setters
        """
        self.lineWidth = 1.0
        self.lineCapStyle = wx.CAP_BUTT
        self.lineJoinStyle = wx.JOIN_MITER
        self.lineDashArray = []
        self.lineDashPhase = 0
        self.miterLimit = None
        self.strokeRGB = wx.Colour(0, 0, 0)
        self.fillRGB = wx.Colour(0, 0, 0)       # used for both shapes & text
        self.fillMode = None

        self.textMatrix = [1, 0, 0, 1, 0, 0]
        self.textLineMatrix = [1, 0, 0, 1, 0, 0]
        self.charSpacing = 0
        self.wordSpacing = 0
        self.horizontalScaling = None
        self.leading = None
        self.font = None
        self.fontSize = None
        self.textRenderMode = None
        self.textRise = 0

#------------------------------------------------------------------------------

class pdfPrintout(wx.Printout):
    """
    Class encapsulating the functionality of printing out the document. The methods below
    over-ride those of the base class and supply document-specific information to the
    printing framework that calls them internally.
    """
    def __init__(self, title, view):
        """
        Pass in the instance of dpViewer to be printed.
        """
        wx.Printout.__init__(self, title)
        self.view = view

    def HasPage(self, pageno):
        """
        Report whether pageno exists.
        """
        if pageno <= self.view.numpages:
            return True
        else:
            return False

    def GetPageInfo(self):
        """
        Supply maximum range of pages and the range to be printed
        These are initial values passed to Printer dialog, where they
        can be amended by user.
        """
        maxnum = self.view.numpages
        return (1, maxnum, 1, maxnum)

    def OnPrintPage(self, page):
        """
        Provide the data for page by rendering the drawing commands
        to the printer DC, MuPDF returns the page content from an internally
        generated bitmap and sfac sets it to a high enough resolution that
        reduces anti-aliasing blur but keeps it small to minimise printing time
        """
        sfac = 1.0
        if mupdf:
            sfac = 4.0
        pageno = page - 1       # zero based
        width = self.view.pagewidth
        height = self.view.pageheight
        self.FitThisSizeToPage(wx.Size(width*sfac, height*sfac))
        dc = self.GetDC()
        gc = wx.GraphicsContext.Create(dc)
        if not mupdf:
            gc.Translate(0, height)
        if wx.PlatformInfo[1] == 'wxMSW' and have_cairo:
            device_scale = wx.ClientDC(self.view).GetPPI()[0]/72.0   # pixels per inch/ppi
            gc.font_scale = 1.0 / device_scale

        self.view.pdfdoc.RenderPage(gc, pageno, sfac)
        return True




#----------------------------------------------------------------------
# This file was generated by ./encode_bitmaps.py
#
from wx.lib.embeddedimage import PyEmbeddedImage

First = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAACz0lEQVRIiX1WsUodQRQ9Z7ME"
    "AoKJxC5EEgyKGk2RP7B8IA/Uwi+wsLS18g8srVKkCQGbZyFJG5tUIoSAIAQtAhaBBBuJmXtS"
    "zNzZ2X1PF5a9b+bOnTtn7jn3ceHzwgcAM7jv0R2vuS3IOmMGwHBWk5wB8EaSSBIAJIkobJEe"
    "VJJopCRBICzGp0UfWfSXCTCgHpntfVnHIMOnsMb24JJQS8phlX54pkVAdTZQCurf6G/Ddk0Q"
    "ACg0UJSwlFB0IIg+TcBs09K4qYDozgscAUUBiaQWPF27khSPHHdUtiUtPl5U/1kfi+OLkkkK"
    "aS59Jx5OaO3lmmQxhoKEgJZdUYTDQSP9eDBw7fkad9/sYn1qnXkuAJUqrr5YxaA3YG+qRwSA"
    "oVgbQAQwQlRAkY9bQuFPiPOz47PYebuDpadLSFWd5zI8RYw6lVyuCkloVUi8H409GMPW6y1s"
    "vNpAxSpXHUUPKi9PCE0VpQGmQVLMlZI20PyTeQ56A0w+moQTsiQmQmetJaIpbZBrvqxrr3UA"
    "0+PTnjFKvvg6BbWydh7kO7irrj1OztRh6Z7AUGadOQEBVetiQqJ5iLaf4OLPBa7/XmPko7Qu"
    "JHkI7cuucrAieHZKG5xcnWDl4wqOzo9Gb1Ak2P1WThInUGl7HUnS1fWVtj9ta/NwE5e/L0vt"
    "koLaJC3syomhoEi0EMmEAEJRqCDQiXT84xj9933uf93Hbbil61QmWsyefg91xtowRJSD7wc4"
    "/XmK81/nrbmbfzfY+7KHw2+HWJ5ebu6ro2GSwLl3cycQltoMTB1qpIDd0b2KDTJZhdOaRsDA"
    "VPdJZtsdCgY2/EBbrpv+0Ei9ZalHnUgCmABjK6vcubqn6WbalfbirZNQJRZqiMlJX1QGyv5K"
    "heT649ij6YCRyamJZzbGgM7GfORCVug+FIc6YP7DAKFGwFkXEplQqOSwlJc+DgcKaLKw4Ow/"
    "oX1dfBQFFGsAAAAASUVORK5CYII=")

#----------------------------------------------------------------------
Prev = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAB3RJTUUH2QYaCSkRaCYhigAA"
    "AAlwSFlzAAALEgAACxIB0t1+/AAAAARnQU1BAACxjwv8YQUAAAMGSURBVHjadVYxS1xBEJ6Z"
    "typYBIsgJBCr6BVJoekMSSXpBDGgpd39AVHzC8Ta2tLmilSmC5gycD8gkEOsxEA0hXAgRsyb"
    "7OzO7Nt9d75j2dndmd2Zb7+ZPex86vQAoQPysW+oTT5qydSsIyIwMSBhGDN6ucKkK7L/DZwf"
    "drzyYtgfmVEssZR9Y2+EQYX8XHNQGnvFoCMHBls/Lw646DizeSqL5qVXkEPighyC5bpGxbks"
    "kYgsttI7Dfdxr1HPrbxHYl/pOjU6uddx3m8e9eMBhUfZHSR8IXo19i4qbGwqXRO/vK346mTS"
    "wg5wKBQFHKTRjUISdeQOWmsCEaNenYQWwoseYYCBVHYqhx7TnLTF54uw9moN22vitPTkSLlB"
    "Y1rVyAEGm/Nt9sksHHw4gOOPx7D0bKlYy+1EDhAlWLKQdVwwZGpiCraWtrj7pgvTE9N6KcAJ"
    "LlKoCaNtYlGVsajheWKIyCsvV3Dn7Q7PzcxZ6kFd10HfwwGJYdSwTgQXMsG8BA0xY8j803nY"
    "e7cHyy+W/RYMI5/aBoJUmuGSC7qPy+8gKFEMcdJNwu77Xdh8vQnhqsTwsQN0M4WmoXPojF6U"
    "sGTB8gEf4PzmnId/h2xwWG8tqwKcUZMT9X1PgSEVRGoqvUSusYbezx6ufl5F30MNdcCUiNCa"
    "JmNDZaU4KmUBswisdiiL2Nh1c3/D+/192Piywf3f/bERFF5rNNaTXLJ5HUqsC/mdkssS52x4"
    "ht1vXdj+vo2Xt5dlBLnXZqv7UUoMTZJwCGGRNGns5dNfp7D+dR0OfxzC7b/bogZZ/cnrlTOe"
    "20RKtCzMXJb1O76Do8ERnFycwMLMwmjtwuYJcBZOSrQqK9dVrOQ23y7XV/dXfP3nuizXunmQ"
    "wUpFXoIzeRyvC732s5o/t/q5/PVR/Bq4qF2us3IOWe3KIbICxQaR0xeNyje2qEuQXrcEQwGF"
    "yfYy1hlEJC9FFeEIL1dWixJc+rrZezvyzyO6MAJPOMDDMLC/IPb3Ixa6cpNiHbJ54IR/qlWN"
    "PPgPVhphQTIBT3gAAAAASUVORK5CYII=")

#----------------------------------------------------------------------
Next = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAACtUlEQVRIiYWWPUtcQRSG33e9"
    "ETGS2EQhWAnGhFik22I7F1KlDegPsPYP+AfEYn+ANnYBCz/QJmAVQqyy2Agq6QRBUiUgiMx5"
    "U8zHnfvhZuByz9w5e2bOM2feWS59XfoCYBGjmp54LNqCrPbNABguC5KLAD5IEkkCgCQRmS0y"
    "BpUkGilJEAjz8WneR+b9ZQIMKFpXO2rVPkgzCyvtGFwSCkkprEInrjQLqNoECkHj2/tb0y4I"
    "AgCFEkWOJUdRQ+B9yoDJpoXvpgzRkxvYgiJDIqmCp253JPmU/YxKtqTeq55mxmcg57/LhTFX"
    "+sNBMh9DToJDxe5QRMRBI2N6MHB5dplH/SOsvVnjBCb8mPNjcAAcGG267LcOhANlQielFNKN"
    "FQDn05wcm8T6+3Xsf9xH/3U/jVX8Ql9OqR/HOjG98EEBU6yUVF1zz+c06A20s7yDhRcLHoUr"
    "scFBoV9FFHHEd8QQDo4vsax1Z7vY+7THje4Gpp9NM8cVEdGRAZ+fIF912mRXzSA2ABjjmFbe"
    "ruDk84lW362qo05jk2GQnFDAgCfrOhy7KCFJOuilYmp8ivMv51GokDljOOHpTEDwE4ys65Ym"
    "CWc3Z9j8tonr39flhgeZoIhyAldOEIMnp5YJbv7cYOv7Fk5/nSbftgMW30WWVq4xpR2w3D/e"
    "Y/vnNnaHu3h4fIhao0z8lMtJmiBWT2UP/CbRy5B0fHXMwY8B7v7epbEa64pcQ2H/rIao8ggY"
    "3g5xcHGA89vzxliFda5bme3l2rUhEmTQ4cVh/JFKrn6sjiLJe24LKGhMiNINZbWUrUxZhqpc"
    "17Akua5lAJgAY6UC/lchkhpY6k8Bl99AzSoKNa48UPJXuASDndgjQwTzupMu8zJgo0Iyvow+"
    "FBs3YPrDAKGAw2UdSVuFZCut+kQcyNAkYcHlP13rmcgeVJ4HAAAAAElFTkSuQmCC")

#----------------------------------------------------------------------
Last = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAACwklEQVRIiX1WPUtcURA9s3lK"
    "bAwoIbAkEEJEsxgSLEw6MWkEC/cn6H/wH2hjbSvEzljZahobUwSEEAgSCRIIa2NjEQMWuXNS"
    "3HvnzX2764NlZ++dNzPnzNfK7KfZjwCmcdfDIR/NMkFtnCkAxXklItMAXpOkiAgAkKTAyRTJ"
    "RklSVIQkQQg02heNOtSoTyWgQDUw2ruijkb6UWgtZ+MkUZE0s0w/cqTOIBsOmIzm76iv/XIl"
    "EAAQoqbC0+KpaFAQdWqDJoumc6WjaGgCB1DhKCFZ0NOUWyQj5OiRJpNcbi+zfb8NhnjOkO4C"
    "OTM+g+7TLucfzpMabTCQCCjkllCQ6RAVyfCgkIVHC3KweIDV56sywpF4F+Ld0pMlbLzdkLUX"
    "a4IASHDvBggChEq0DFKCmysAIcIcuzeG9Zfr2H+/j7nJueLOqI1Rg4HxLsD0WhleOmCiKSYT"
    "YtU19WCKu+92uflmExOjE3QFUL8bSIaSriolRJKiCMUqJUVpDQgAK89WuPh4UXp/evE8Iijf"
    "1dRoTA6s5n1d51r3/ZGe8dFxdiY7ds5AGBLXB7FMFRha1wMQ5JFioyUF5qK2ngBRJjknyxJV"
    "xF0/t/9ucXF9USY5pPEQXF8o0DJjzrgpDXBw8vsE3b0ujn8d1w5cgM3vysHyMybKjparv1fY"
    "+ryFo59H8M3Yl4NyVqUqUpQ50DoHSuXe9z3Z/rKNm9sbf9eXA2dcagehAdPNktPLU+yc7uDs"
    "6qzv7vDHIXrXPVxeX9b5aswwkpDOh85XEK/KIZU21MABNmR7OQek7YtvlagYRbahmpC1hkxF"
    "Oa4btNi4Zl44gRYZorM6kry5mmiakTZHu/tUCH4Dsa+T0/CjN2T6TIWUZOMe9QaMnZyWuHVj"
    "NJi70SC7sSJZRyh9G9D+MICoEHDepIRKCKVc5HWkpU6mA44aGyw4/w8M23JBubTHcwAAAABJ"
    "RU5ErkJggg==")

#----------------------------------------------------------------------
PrintIt = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAADDUlEQVRIiZ2Wz2pUWRDGf9Xc"
    "yAgRxkVDFsFGRARHISADbdCNvkCEPIAL8R3c+RZKnmBmp45usuyFzKySxRCT+GehkUQik0Fb"
    "O9fuez4X554/99rJwgPNrXO6uk7VV9/9qu3S6qU/gAsct3TExwVbyLXOHODYKszsArAgSWZm"
    "AJJkZLbMQlBJMmcmSQjD+fjmvI+c95cTOCimZntc1j7Ij1W4ZIfgkigkxbCqNyHTLKBaF6gO"
    "Gp7e3/1oF4YBmEhQBFgWfl3geve6h8InQB3SEELY+sd1Bu8HwgW4PIQ4X0mCaEoDb83fYml+"
    "ibo1hEvMLD7/2fubwdvBVKhw1BD5bJqw+CpCUJGtsPdPC5gfAZEMhMXS0kU2dmOAyK7IKvMs"
    "MjMbV2OomizK4KJIXK4xdlBQgODBxgNW3642GVT7hP27T++Y0QwA42o8BSKXIEJopb9Cv9vP"
    "Cdu222SOZ893nnP3r7uhCnDQqQ3DQfdE1/rdvqdVvXI730/zWZxftLmTc1hlRuVjFrEpgo46"
    "MaO8scc3uXk2w0ysQE6+AnNmVDAshyb00xUA9qX8AhWxgk5sioOv5VdGkxE/uypVDEdD8pgF"
    "FbHjTo790T69Uz3WPqzx7M2zJnuO0Kab526y2Fvk8+Fnym/lESzymOnlwUt6p3o8ff1U5aRk"
    "6fwSBLYk7kRr8GbAo38fcfXMVW182EgvXbggsCjI7L3BPR7OPtSr/17Z/Wv3uTJ3RU+2ntjB"
    "6CAFFzZ7YlbLl5dtMplw58872t7btp2DHaj89+mCDCIcHI4P2RxtJgiAG2dvUL/V8azoJBmr"
    "JhXbe9utIRTkumpC5J2EXALi8YvH2h/u+8lSg3X6l9Pc/v12U67btqAwZxGiOKGc15bdT7uU"
    "k1LLvy0nOaWWa1A5KW33/92AeZL6pGnYxZWLa4gFnETQ8TAWEwEsm1ZhJOZYNy9Ic2W9oMrl"
    "NSuzttusiGrrbWXil6SeDKLwJjeyDhnVPwwlZ/ha8DFZO+v0hwFRULGVZR2Htsmag7wl09En"
    "Y1u0U8e2vgN1UBiOxfpC0gAAAABJRU5ErkJggg==")

#----------------------------------------------------------------------
SaveIt = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAADAElEQVRIiZ1Wv09TURT+zuuj"
    "JWmMOomOrQaDDCTGri7MQoh/RfkLmFmY2AqJA7MbDQODjP5IQ2JEJwiJSNKURBY1OLT0ns/h"
    "/nwPJMabvPTc+84797vf+c65ldk3s68BTOOmwb886m2CWlpTAIqjXESmAcyRpIgIAJCkILEp"
    "4oOSpKgISYIQqI0van2o1p9KQIH8WrQ3obZBrp5Co+2Dk0ROMoSlm3ikSUCWNqAL6n+tv161"
    "c4EAgBD2aFO1Kczfm7cU2YAEIG5TOyfEvQk+JGXvdA9nF2eWLrUniRQ5dAsPFtB+2IZLB0gG"
    "O52n696uZ3VsHmwWaMsTVCDJDJn/iEnQYKfzso9QAFOmiAIQ4tXhksyRjmT7dBsjMwoU+Xfe"
    "H4RUsyqWHi2xVqmJUGxe1KuLyKOW7YIH1f/dx9rnNYzNGO5D+y4mEqKCCipoTbXQvNO06jEo"
    "UaSRoqAKAI1bDfZe9DDWcSpeDyCIeSKbQK1SC6qjcfE8RW43cYsFirpfuxiZka27KISCiqpZ"
    "FUuPLUUepKgrNLoNguaTE/Qv+lz7aClKdO+f4F9BBa37LTTuNizqJMlWpgq40o4tAWDzdlN6"
    "L3u41Eub5EREbk4A4iiybcaDdEkOJyihC2Myn0SNtX+qA781DYMo7AYmbuBq4v9HqScVVaRJ"
    "j7Go2D3uYvXdKoaXQ/oGliQS9Yk6V56vYPHJImgHCnGCihQxB5YiiogMfg0wHA3Zftq+Umgg"
    "2HnfkcHPQfAXFcAkPUpLFEGBkx8nhSPTEMvPlq/NQedtp0hp4RKyvhkNQUPCADTk7vEutw62"
    "vHTpZczAQdKLDH2JgiSVGqimWqnmohIo8jfU+od1Hn4/lPOLc4gRbvQ2IkVJoYmK7H/b9wFl"
    "58tOuPH8CWTm1cwnEHNQEr6P+2sxCiDoOrkSU66DnV6xJA/yYnuN2fdKcOphym/wJ4JyQj+z"
    "mIJtK9ld4gG1R+Q+DG08thXxPkIpo45/GEDkMDhKUIdLO1RjRF1UiPeJ8o12LPyjPzWYSc7N"
    "KY0aAAAAAElFTkSuQmCC")

#----------------------------------------------------------------------
Left = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAACzUlEQVRIiZ1WPUsdURA9Z11B"
    "7EMIgRQJxCJGEkgidoKFpWATK3+AKFgIgo1YiSAIr4hia5N/IKRRCSSFwSbVS5MukB+g+JE7"
    "J8X9fk8tXLjs7N3ZuTNnZs4sR7+MfgYwgvsu3bEsyoKsZ88AGLotyREAbySJJAFAkohCFhmN"
    "ShKNlCQIhHn7NK8j8/oyAQa0t3p7n9feSH8UluVoXBJaScmswkP0tDCongMUjMa717d+uSGI"
    "CAdF0kgYQMtygIJTj6dII2gkXNBxQceBcKBMoAv7DmjuTKTVa3FkETvvd0Bj3ncBDufluJdk"
    "Q4BIPtQKlhimgJVXK5p/Me8dMUDO68m8jiQP1y0QtRQBgbE6ioNIEWuv1zT3fI4pUgfBwGCQ"
    "MRdlFYX3Psm5lpMngIBGDTbebmDm2UxdZAESSRVc6duimjxEliGCeecHMYjNd5uafjqNqroA"
    "nH086y3mdPbuj13snu56exGiVCV+k0PNELY/bGvyyWRutNCAADCAAZG+0arGJNmgSXDFXmlT"
    "UgQMN8PqTHQw/mi88rqUy+c+nVhZQZaFHITWVoPGewH0eVcauiuC1HwhyRDQlIk5vzrHwvEC"
    "Tv6c4EGXALk62U1qjNA0l9eXWD5axuHvwwcdkJot3HMVWeaYG7vB6vGqLq4vMPtytsJ67NMY"
    "nHOxSmLD5cqpuSpA5EA5ZW5xgPvnuH6yzoOfBwDBeIVoGQ5glOkCd7mwH/LQlLxRwhUoAVtf"
    "t7B/uo8UhOV31XfWvzxdu36IYIIsh9z51tHVzRWWJpYgp5qLSrrulQW0od0ZlD2H9HCLTNz7"
    "vofu3y7kFBspTbr0bTHxIvXECAATELklehInVwj56NcRZALFzD/VbO5fbdl5KEM25QopJlcP"
    "lasgyEz1KCCK0ysN82wwdmMKucCXUYdi9WNQ/TBAaOHQ7YWkhKHwuoIi6UQ4UECTp0f3P4CK"
    "ICuf1hoBAAAAAElFTkSuQmCC")

#----------------------------------------------------------------------
Right = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAACs0lEQVRIiZ2Wv0scQRTHv9+7"
    "BU//gZSBELwEhaQIaS0sLNJYWlnGC1xlo2ChrSAIVwhXWySNljFYnEIkpDLEVJdA/gohgs77"
    "ptiZ2ZndnGAGhns7vHv73uf9mOX86fwHAF3ctzRhW5AFWe3MABjGBckugJeSRJIAIElEIosM"
    "RiWJRkoSBMJK+7RSR1bqywQYUPzT2/u8Lo00o7BKDsYloZAUzco/BE8Tg6q9QN5o+C31rSm3"
    "CCLgoEgaCQNolexRkEYuPloEjYTzOs7rOBAOlAl0/twBrYmJtOamiP3X++g/74NGwHkcDuU2"
    "5LLBI1IZaoYlhFlHBODts7fqsIO9yz0F1jBMQCQiVEmGxUosMqXnJUqSq91VbL3aIuVxueS/"
    "Ca6iquXoSbM6wm9trcyuYIpT2L7YhjOXI1VAZBUiGLQ2u4Zet5cWbF2O1SZJy0+XMd2axsZo"
    "Q7fuNhiPiEIE9IdsoYU223nTeTl9Ts+Xniyp0+5w/dM6buymbETfKy1YmUSZQjKjd6mn9ee6"
    "zsLjBR28OdBMa6a0Y5Ccz4Fv7WokAA+KIMgEQVFwCP1TRhB3qN//WOe/z9E77uH6z3WW7AKu"
    "yrjviQevk/EJNj9u4u7urlF5nHs/9w3CC5+LZheHl5vQVhtX764y40c/jrBzulOVaf6C70Vo"
    "qpgDn3QYGAdaOYeCzZiDw8tD7o52YWbZuPZV2UQ0yXOK1RlKlMOvQww+D3J95XI5rp1iY1Rj"
    "WFCYLYoRhQrG4GKg4ZdhOrarOZbKAgoaI6J4Q1l+QwVcktA/7uvs11lW1imWcOPVIwBMgLGB"
    "J8UnCaOfo4gsVl2KqLYLuHS8JmF62V9/Sg0lo1zJgKxGPRJE4faKl3llMHRjDDnhy6BDMfsw"
    "yD4YIBRwGNeRZJVTeZ1XSNBJLqIox4rG+C+fohr4fgXssAAAAABJRU5ErkJggg==")

#----------------------------------------------------------------------
Width = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAADeUlEQVRIiZWWT2gVVxTGf2cy"
    "L5TEQDA2aUFINmKJjyiUtjQidGHblcsGQRBx1Y3BjRtFXIkLF4ILcRehWmtWQRfWUIpdJC5E"
    "07hJUpAUJWmLpf5pCYS8ez8Xd+78y4vQgeGdmTnz3Xu+Od93njWnmz8Au3nXoS1OH2MhX7vn"
    "Ac9Sama7gX2SZGYGIElGKZZZBJUk82aShDB8wDcfcuRDvrzAQ9p2t+/adQDZXIUv4gguiVRS"
    "DqvsIu60BKjaAspA42/I95vj1DAAEwUVZVrKVNQoCDkFYB6bz+57lSja8gO2oaJEiaQKPfU4"
    "kRRKDisqj1XEPR09Gn1/VHLhen//fnUn3eE9h+QDhpyEoxInJiPSYd4slpeVbCO9I0x+MWkH"
    "PzwYnjk4NHjIbn11i+HeYcNjODBXetdhOExeJHlJWbmxA8wbx3cd59qBa+zs3lnpDgSDPYNc"
    "//o6Rz86SuIT5IWcwBHOsBBJLC+7IXmpr9HH1c+v6uTwSaWWxg8d6HAi6x0aSUOnPjmlK19e"
    "YXvndslJclW6rHm7OYfYhxAeG+0f5fzH57XjvR2F0Mxs5b8VFl8ugtCevj020DVQEeaLtRd2"
    "+qfTzD6bDUIMWvnVmlPNOUl7U1LGh8c5tusYWev+78PLM/F4gsszl2m1WshrPsn4tk46Geoe"
    "ssSSIIzsKMfl63Y5HUmHDfUO0aBhuKCLjv5v+r9FfLDR2uDes3v2av0lnw58RpqkeRlmZiv/"
    "rvDwj4csv1qmq9Fl2zq35QsBrLt1u/DzBS7ev0ir1YrC/DPJv7oD7zw3Fr7nyN0jLL9erpQ/"
    "83yG8R/HOXH3BI9WH1WePf3nKYdvHObm45tFJ7XrIrkgroW/Fxi7Paap36YKjxK50GIXSdLk"
    "/KTGvhtj6a8ltRNsGkVV8RYvrW2s2Zn7Z3jw/IHOHjgbhRiNzt6sv9G56XM2vThd+E9hhBat"
    "IsXVPKTmJXcW7/Bk9Qkj/SN56bO/z3Lpl0usvl6t5tc8TBI2PDE8h9hbBc4mVFsD22J6lRaQ"
    "8nkxn5q3nKJ8QvnqhMIXJctTtesaLRW6JNJM+uAFYbFiJ3Fy1aup77Ru7aUzxZUnkFQCV8ng"
    "VAbK85UNwdhZkXuKCZhGJefDvACMYslLzqdXeCaEmWzTBMz/MCBSHEt1SuSFyaqDvNhpNSfS"
    "QYmawsqW3gIVgl7eufygTQAAAABJRU5ErkJggg==")

#----------------------------------------------------------------------
Height = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAADXklEQVRIiX1Wz2tcVRT+vjfP"
    "dKE4jNhBI1kISogMxlQZQgKW/gtdtnQZaLclq5Z2kRYyJH9DtyXQLkSxopBBitK6KKmIDBmD"
    "CxeGGLEEIWli7/lc3B/vvb6ZXri8M/eec+493z3fOcPOd50NANN41dCYaVEWZC+tGQDDdk5y"
    "GsAnkkSSACBJREkWGZ1KEo2UJAiEef80ryPz+jIBBuQjbzvm1g00AAEv7EU9Civk6FwScknJ"
    "rcKPeNN0a/Py1Y+uQhLWfllTcBq/Xt/qck4QACgUUJRhiVCcbZ/FxfcvAoAe7z3mwz8fep3C"
    "YZJpYd2EbOxDWjHbE22szK4gYwaCuP3pbbRPtQEX4HDw01CVDcgkeQj8iUqyvJwpw+rcqloT"
    "LUUIWxMtrXZXQVFwkMz7kJPCoUnOKCLCQSNjeCFkLn24hO7pLuMAAJLstrtYmlkiDIQD6Eq2"
    "DoQDPUQlKEI0KdQzb53B5enLY+lx5eMrmHt7LkElV4cri+GFBQWY0Myb6n3WU4MND1cY5Wxr"
    "sKG1z9fUfK3pbZ0kV4UrD7lLCIKBFCGTFt5d4OCfAQYYSBJnWjOYfGMSknRwfMAne08U7LA4"
    "uagHwwfJlhaIJoCdLzpbkmbHkSam28r8Cs5/cB4AsLW3hUtfXUp7r7D7OYcBgdqj89oKTpCM"
    "cBEuRGwsyoe/dbKFCogqDw0Dzr13DvPvzCdezJ6eTY879eYUri1eS3uP/niE/m/9VCYoojjA"
    "lUIKzmXCYH+AW/O30DzVRMjONNqvt3GhcwEk8ezoGe78dKcGT/xWskiuINruv7u6+cPNlDWj"
    "skiSbnx7Q7sHu1WSluQsEkNOnmiODLnMzd83ufHrBlgakWgAcPfpXfaHfcJKRHMFSf0BpbqR"
    "SFIiyvqP6xj+PayRbLg/xHp/vbCz+pSETE5INSRB5NfkpOcnz7H8zbKO/jtKEB2eHGr5y2Uc"
    "nxyPhKUCEY1IsFioIa4a8s7+Dnvf9wj42t7b7HHnr52RtUtSZT2XU3hxAcYawaJ87+k9LEwt"
    "QBLub90vsq7Sm+szhyt3IKnkXKX2Jwi4/vX1Ys87VUxtCEUHRNENPZNDE09s9A4jG1NDPzw6"
    "TB0u6lCsdcD0hwFCDoftlyGpsDE28uKmVZ0IB0rQFLzc/h/4UISz4M1LhAAAAABJRU5ErkJg"
    "gg==")

#----------------------------------------------------------------------
ZoomIn = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAELklEQVRIiX1WTWhUVxg95817"
    "GlON2ug0KRiJaJOQBEWarm0F46qLGqhIcZ1EkEYCulBcunAhCLrJSgraRaG0hSyiEXUlGcmk"
    "Tn9MLTb+jEyiScAkxsx47+nivfvmzUQdeMz37vveufc75/t57Bjp+BFACz7003su62xBtmrN"
    "ArCY9Em2ANgjSSQJAJJEJGyRDlSSaElJgkDYEJ829JEN/WUFWMB/52k/dOoQZHUUtmw7cEnw"
    "JcWwim7cSROAqtpAEaj7D/3tatsnCAAUylQ4Wnau34nuhm41fdTEoini/tx9DD8d1uLbRToq"
    "EoCxTRutW4GdI53ZWINoAw+eBlsHeWT7ERCMtQGAuZU5nc6c5p3nd5wWYSQWZQ3Kekyk0t+l"
    "eyF8kqTgVNspROC4VbiFq4+uYmJ2Ag3rGtCwrgHd27pxb/oeni8+j7mONajUZNqjwjNCAC3Z"
    "ur6Vh7cfBgCey53j8bvHsXnNZmZfZtkz0sOxmTEEXsCzn59FRAVhABrSUQMDwoCygpfcWRK6"
    "G7tBELn5HK79ew0yQn9bP7rqu7BcWsaZu2dgrEFzXTP2btkLGAAm5FtG8X20ETxZhVkRLmhb"
    "7TYBwPjLcfW19il3KAdJ6mvvU+7bnPKv8ni2+EwA0L65XTLld2UkmRBPRoIB/IgzQqFQS6Ul"
    "AFD92nrezt8GDNTf2c9MIYPMdAYBAm1au4kAFDBgFL1gQYqxyK5WvCid5CLJzGQEAPs/3a/8"
    "Ql6X71+GJI0VxnQpe0k9O3tQt6ZOAPDf/KN3njpiRDKRBrQMhbLkyJMRPl54jNqglkP7hti1"
    "tQu7f9jNK39c4dG2ozz5xUm4tJ1ZfEGaUFgaEgYVIsMC7PipIwthTxyahF11uzT01RC3rNsC"
    "SVoxK/Q9H77nw/UsSXq18orHfjuG8fx4ZR2UC3AilT6U7oVFQ6yFBWZfz2L40TA3BBvQVNeE"
    "2qCWHj28tW+RX8hjY81GAkBNUMODnx3EX4U/8WT+KV0NuEKDRYHt19qzEHa70q5oXBZIIYX6"
    "mnrQEvPL85ARLhy8gH3N++L+WDIlnPj1BEb/Ga1ugL+n0t+ke2HRKKuwKbhiCUOmjLD0ZolL"
    "b5ZojKExBjce3mDr1hY0f7yDJOnRw4GWA3ww/YBTL6ZcVhIWBS8uiqgwqgtFRqueF0tFfP/L"
    "AG4+vBlHEaQCDH45WBG9JHgyQpxeRgpTLFxzReTs5LNiqYiBnwc0Ojkat/rJ6Ukl0l6wQCr9"
    "dbqXlo0JkakodV02OLriZ9G6NZbX/75OgpianeL5kfN8XXxdflco+DJOWAGWFd1wlehVE0wS"
    "Sirh4s2L751+PkxyAkkJcCXGn5Lgsb+iIRjZ8cRDeQL6rpLjYV4GdHmtRF67UUrnQ7Hiw6Di"
    "gwGCD4PJakqilK0c5OWTVvo4OpCgJp5/mPwf/ayV5rr3q+AAAAAASUVORK5CYII=")

#----------------------------------------------------------------------
ZoomOut = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAEE0lEQVRIiX1WTUxUVxT+vjdv"
    "Kj8FbcERmoCh0Q4ESBsT66ZNaE2EVY3tStO4hrgphkQXunbhwkTdsW2kJk2a1oQ0KERdNYEA"
    "StpKbSkqENACqYLIDPd+Xbx737wZqpNM3nn3nXfuud/5vnMe24bavgOQxZt+es3feluQLVmz"
    "ACymQ5JZAB9JEkkCgCQRCVukDypJtKQkQSBsFJ828pGN/GUFWCD832zflHUUZPspbMH2wSUh"
    "lBSHlbvxmSYCqmQDuaD+Gvnb7XZIEAAoFKDwsOx7ex866zrVWNnInMnh/sp9DD4Z1NrWGj0U"
    "iYCxTevWrcD2ofaJuAZugwCB+pr7eGLvCRCMawMAK5srOjd6jncX7vpaRCexKNSgUI/JVObr"
    "TDeEPUkIzrachQuO24u3cW3mGiaXJ1FXXoe68jp0NnRibGkMC2sLMdZxDYprssT2n9snIHcC"
    "S2arsrj+6XUR5IWpCxj4ayDOroxluPrJVR3ac4gz/87o6I2jlFVp1km4JoPkzpLQWd8Jgpha"
    "ncLAnwOQcdkZYCO/gfO/nIexBk3VTThQewAw0TNZRb7u3r8TOsrFrGioaAAAjP8zrp7mHvS0"
    "9CTJi7Zv2zC3NqfGqka0vtOqsbkxD4k8PSEUWOQW6Ba5nl8HANXsqOGd+TuAQSQot00aae3a"
    "sYsAlGaaLpBgQYoxXF4rYUwzl8Xo01Ec23sMh987rCv3rmB0YTTJexzPHkf1W9UCgL9XZxRD"
    "mMja60BWCGABWhImug49HuKjF49Qka5gf0c/D+4+iJRSrAwqebLlJM98fAaetk/XnpEm4jwN"
    "CRMVFwaxzbbv2yIW2bjHYH/1fvV/3s/a8lpI0qbZZBiECIMQvmdJ0vPN5zx14xTG58eLdVAQ"
    "4GQq81WmGxZ1cS0ssPxyGYMzg6xKV6GxuhEV6QoGDLBltzD/Yh47y3YSAMrSZez6oAu/Lf6K"
    "x6tP6DXgKQuLRbYOtE5A+NBjVtS4LJBCCjVlNaAlVjdWISNc6rqEjqaOuD/mTR6nfzqN4T+G"
    "SxvgvVTmy0w3LOplFTUF10fckSkjrL9a5/qrdRpjaIzBrYe32Lw7i6Z33ydJBgxwJHuED5Ye"
    "cPbZrGclYbEYxKJwwigVSsySxPNcPodvfuzFyMOR+BTpVBp9n/UVnV4SAhlBRnLBFEk9WpNx"
    "dHN28lkun0PvD70anh6OW/300rSnp/MFUpkvMt20rE8UmXLU9WzwcMXP3Lo1ljd/v0mCmF2e"
    "5cWhi3yZe1l4V1gMZXxhBVgWdcNtRS+ZYJKQVx6XRy6/dvqFMMkJVFCsV6/rL0oGj/3lhqBX"
    "sp94KEzA0Cs5HuaFgJ7XSvDatxV6H4pFHwZFHwwQQhhMl0LiKFs8yAuZFvt4OJCAJp5/mP4P"
    "yFN2/NjDF0IAAAAASUVORK5CYII=")




"""
This module supports the thread-safe, asynchronous transmission of data
('delayed results') from a worker (non-GUI) thread to the main thread. Ie you don't
need to mutex lock any data, the worker thread doesn't wait (or even check)
for the result to be received, and the main thread doesn't wait for the
worker thread to send the result. Instead, the consumer will be called
automatically by the wx app when the worker thread result is available.

In most cases you just need to use startWorker() with the correct parameters
(your worker function and your 'consumer' in the simplest of cases). The
only requirement on consumer is that it must accept a DelayedResult instance
as first arg.

In the following example, this will call consumer(delayedResult) with the
return value from workerFn::

    from delayedresult import startWorker
    startWorker(consumer, workerFn)

More advanced uses:

- The other parameters to startWorker()
- Derive from Producer to override _extraInfo (e.g. to provide traceback info)
- Create your own worker-function-thread wrapper instead of using Producer
- Create your own Handler-like wrapper to pre- or post-process the result
  (see PreProcessChain)
- Derive from Sender to use your own way of making result hop over the
  "thread boundary" (from non-main thread to main thread), e.g. using Queue

Thanks to Josiah Carlson for critical feedback/ideas that helped me
improve this module.

:Copyright: (c) 2006 by Oliver Schoenborn
:License: wxWidgets license
:Version: 1.0

"""

__author__  = 'Oliver Schoenborn at utoronto dot ca'
__version__ = '1.0'

__all__ = ('Sender', 'SenderNoWx', 'SenderWxEvent', 'SenderCallAfter',
    'Handler', 'DelayedResult', 'Producer', 'startWorker', 'PreProcessChain')


import wx
import threading
import traceback


class Struct:
    """
    An object that has attributes built from the dictionary given in
    constructor. So ss=Struct(a=1, b='b') will satisfy assert ss.a == 1
    and assert ss.b == 'b'.
    """

    def __init__(self, **kwargs):
        self.__dict__.update( kwargs )


class Handler:
    """
    Bind some of the arguments and keyword arguments of a callable ('listener').
    Then when the Handler instance is called (e.g. `handler(result, **kwargs)`)
    the result is passed as first argument to callable, the kwargs is
    combined with those given at construction, and the args are those
    given at construction. Its return value is returned.
    """
    def __init__(self, listener, *args, **kwargs ):
        """Bind args and kwargs to listener. """
        self.__listener = listener
        self.__args = args
        self.__kwargs = kwargs

    def __call__(self, result, **moreKwargs):
        """Listener is assumed to take result as first `arg`, then `*args`,
        then the combination of moreKwargs and the kwargs given at construction."""
        if moreKwargs:
            moreKwargs.update(self.__kwargs)
        else:
            moreKwargs = self.__kwargs
        return self.__listener(result, *self.__args, **moreKwargs)


class Sender:
    """
    Base class for various kinds of senders. A sender sends a result
    produced by a worker funtion to a result handler (listener). Note
    that each sender can be given a "job id". This can be anything
    (number, string, id, and object, etc) and is not used, it is
    simply added as attribute whenever a DelayedResult is created.
    This allows you to know, if desired, what result corresponds to
    which sender. Note that uniqueness is not necessary.

    Derive from this class if none of the existing derived classes
    are adequate, and override _sendImpl().
    """

    def __init__(self, jobID=None):
        """The optional jobID can be anything that you want to use to
        track which sender particular results come from. """
        self.__jobID = jobID

    def getJobID(self):
        """Return the jobID given at construction"""
        return self.__jobID

    def sendResult(self, result):
        """This will send the result to handler, using whatever
        technique the derived class uses. """
        delayedResult = DelayedResult(result, jobID=self.__jobID)
        self._sendImpl(delayedResult)

    def sendException(self, exception, extraInfo = None, originalTb = None):
        """Use this when the worker function raised an exception.
        The *exception* is the instance of Exception caught. The extraInfo
        could be anything you want (e.g. locals or traceback etc),
        it will be added to the exception as attribute 'extraInfo'. The
        exception will be raised when DelayedResult.get() is called."""
        assert exception is not None
        delayedResult = DelayedResult(extraInfo,
            exception=exception, jobID=self.__jobID, originalTb=originalTb)
        self._sendImpl(delayedResult)

    def _sendImpl(self, delayedResult):
        msg = '_sendImpl() must be implemented in %s' % self.__class__
        raise NotImplementedError(msg)


class SenderNoWx( Sender ):
    """
    Sender that works without wx. The results are sent directly, ie
    the consumer will get them "in the worker thread". So it should
    only be used for testing.
    """
    def __init__(self, consumer, jobID=None, args=(), kwargs={}):
        """The consumer can be any callable of the form
        `callable(result, *args, **kwargs)`"""
        Sender.__init__(self, jobID)
        if args or kwargs:
            self.__consumer = Handler(consumer, *args, **kwargs)
        else:
            self.__consumer = consumer

    def _sendImpl(self, delayedResult):
        self.__consumer(delayedResult)


class SenderWxEvent( Sender ):
    """
    This sender sends the delayed result produced in the worker thread
    to an event handler in the main thread, via a wx event of class
    *eventClass*. The result is an attribute of the event (default:
    "delayedResult".
    """
    def __init__(self, handler, eventClass, resultAttr="delayedResult",
        jobID=None, **kwargs):
        """The handler must derive from wx.EvtHandler. The event class
        is typically the first item in the pair returned by
        wx.lib.newevent.NewEvent(). You can use the *resultAttr*
        to change the attribute name of the generated event's
        delayed result. """
        Sender.__init__(self, jobID)
        if not isinstance(handler, wx.EvtHandler):
            msg = 'SenderWxEvent(handler=%s, ...) not allowed,' % type(handler)
            msg = '%s handler must derive from wx.EvtHandler' % msg
            raise ValueError(msg)
        self.__consumer = Struct(handler=handler, eventClass=eventClass,
                                 resultAttr=resultAttr, kwargs=kwargs)

    def _sendImpl(self, delayedResult):
        """Must not modify the consumer (that was created at construction)
        since might be shared by several senders, each sending from
        separate threads."""
        consumer = self.__consumer
        kwargs = consumer.kwargs.copy()
        kwargs[ consumer.resultAttr ] = delayedResult
        event = consumer.eventClass(** kwargs)
        wx.PostEvent(consumer.handler, event)


class SenderCallAfter( Sender ):
    """
    This sender sends the delayed result produced in the worker thread
    to a callable in the main thread, via wx.CallAfter.
    """
    def __init__(self, listener, jobID=None, args=(), kwargs={}):
        Sender.__init__(self, jobID)
        if args or kwargs:
            self.__consumer = Handler(listener, *args, **kwargs)
        else:
            self.__consumer = listener

    def _sendImpl(self, delayedResult):
        wx.CallAfter(self.__consumer, delayedResult)


class DelayedResult:
    """
    Represent the actual delayed result coming from the non-main thread.
    An instance of this is given to the result handler. This result is
    either a (reference to a) the value sent, or an exception.
    If the latter, the exception is raised when the get() method gets
    called.
    """

    def __init__(self, result, jobID=None, exception = None, originalTb = None):
        """You should never have to call this yourself. A DelayedResult
        is created by a concrete Sender for you."""
        self.__result = result
        self.__exception = exception
        self.__original_traceback = originalTb
        self.__jobID = jobID

    def getJobID(self):
        """Return the jobID given when Sender initialized,
        or None if none given. """
        return self.__jobID

    def get(self):
        """Get the result. If an exception was sent instead of a result,
        (via Sender's sendExcept()), that **exception is raised**, and
        the original traceback is available as the 'originalTraceback'
        variable in the exception object.

        Otherwise, the result is simply returned.
        """
        if self.__exception: # exception was raised!
            self.__exception.extraInfo = self.__result
            self.__exception.originalTraceback = self.__original_traceback
            raise self.__exception

        return self.__result


class AbortedException(Exception):
    """Raise this in your worker function so that the sender knows
    not to send a result to handler."""
    pass


class Producer(threading.Thread):
    """
    Represent the worker thread that produces delayed results.
    It causes the given function to run in a separate thread,
    and a sender to be used to send the return value of the function.
    As with any threading.Thread, instantiate and call start().
    Note that if the workerFn raises AbortedException, the result is not
    sent and the thread terminates gracefully.
    """

    def __init__(self, sender, workerFn, args=(), kwargs={},
                 name=None, group=None, daemon=False,
                 sendReturn=True, senderArg=None):
        """The sender will send the return value of
        `workerFn(*args, **kwargs)` to the main thread. The name and group
        are same as threading.Thread constructor parameters. Daemon causes
        setDaemon() to be called. If sendReturn is False, then the return
        value of workerFn() will not be sent. If senderArg is given, it
        must be the name of the keyword arg to use to pass the sender into
        the workerFn, so the function can send (typically many) results."""
        if senderArg:
            kwargs[senderArg] = sender
        def wrapper():
            try:
                result = workerFn(*args, **kwargs)
            except AbortedException:
                pass
            except Exception as exc:
                originalTb = traceback.format_exc()
                extraInfo = self._extraInfo(exc)
                sender.sendException(exc, extraInfo, originalTb)
            else:
                if sendReturn:
                    sender.sendResult(result)

        threading.Thread.__init__(self, name=name, group=group, target=wrapper)
        if daemon:
            self.setDaemon(daemon)

    def _extraInfo(self, exception):
        """This method could be overridden in a derived class to provide
        extra information when an exception is being sent instead of a
        result. """
        return None


class AbortEvent:
    """
    Convenience class that represents a kind of threading.Event that
    raises AbortedException when called (see the __call__ method, everything
    else is just to make it look like threading.Event).
    """

    def __init__(self):
        self.__ev = threading.Event()

    def __call__(self, timeout=None):
        """See if event has been set (wait at most timeout if given).  If so,
        raise AbortedException. Otherwise return None. Allows you to do
        'while not event():' which will always succeed unless the event
        has been set (then AbortedException will cause while to exit)."""
        if timeout:
            self.__ev.wait(timeout)
        if self.__ev.isSet():
            raise AbortedException()
        return None

    def __getattr__(self, name):
        """This allows us to be a kind of threading.Event."""
        if name in ('set','clear','wait','isSet'):
            return getattr(self.__ev, name)


def startWorker(
    consumer, workerFn,
    cargs=(), ckwargs={},
    wargs=(), wkwargs={},
    jobID=None, group=None, daemon=False,
    sendReturn=True, senderArg=None):
    """
    Convenience function to send data produced by `workerFn(*wargs, **wkwargs)`
    running in separate thread, to a `consumer(*cargs, **ckwargs)` running in
    the main thread. This function merely creates a SenderCallAfter (or a
    SenderWxEvent, if consumer derives from wx.EvtHandler), and a Producer,
    and returns immediately after starting the Producer thread. The jobID
    is used for the Sender and as name for the Producer thread. Returns the
    thread created, in case caller needs join/etc.
    """

    if isinstance(consumer, wx.EvtHandler):
        eventClass = cargs[0]
        sender = SenderWxEvent(consumer, eventClass, jobID=jobID, **ckwargs)
    else:
        sender = SenderCallAfter(consumer, jobID, args=cargs, kwargs=ckwargs)

    thread = Producer(
        sender, workerFn, args=wargs, kwargs=wkwargs,
        name=jobID, group=group, daemon=daemon,
        senderArg=senderArg, sendReturn=sendReturn)

    thread.start()
    return thread


class PreProcessChain:
    """
    Represent a 'delayed result pre-processing chain', a kind of Handler.
    Useful when lower-level objects need to apply a sequence of transformations
    to the delayed result before handing it over to a final handler.
    This allows the starter of the worker function to not know
    anything about the lower-level objects.
    """
    def __init__(self, handler, *args, **kwargs):
        """Wrap `handler(result, *args, **kwargs)` so that the result
        it receives has been transformed by us. """
        if handler is None:# assume rhs is a chain
            self.__chain = args[0]
        else:
            if args or kwargs:
                handler = Handler(handler, *args, **kwargs)
            self.__chain = [handler]

    def addSub(self, callable, *args, **kwargs):
        """Add a sub-callable, ie a `callable(result, *args, **kwargs)`
        that returns a transformed result to the previously added
        sub-callable (or the handler given at construction, if this is
        the first call to addSub). """
        self.__chain.append( Handler(callable, *args, **kwargs) )

    def clone(self):
        """Clone the chain. Shallow only. Useful when several threads
        must be started but have different sub-callables. """
        return PreProcessChain(None, self.__chain[:] )

    def cloneAddSub(self, callable, *args, **kwargs):
        """Convenience method that first clones self, then calls addSub()
        on that clone with given arguments. """
        cc = self.clone()
        cc.addSub(callable, *args, **kwargs)

    def count(self):
        """How many pre-processors in the chain"""
        return len(self.__chain)

    class Traverser:
        """
        Traverses the chain of pre-processors it is given, transforming
        the original delayedResult along the way. The return value of each
        callable added via addSub() is given to the previous addSub() callable,
        until the handler is reached.
        """
        def __init__(self, delayedResult, chain):
            self.__dr = delayedResult
            self.__chain = chain

        def get(self):
            """This makes handler think we are a delayedResult."""
            if not self.__chain:
                return self.__dr.get()

            handler = self.__chain[0]
            del self.__chain[0]
            return handler(self)

        def getJobID(self):
            """Return the job id for the delayedResult we transform."""
            return self.__dr.getJobID()


    def __call__(self, delayedResult):
        """This makes us a Handler. We just call handler(Traverser). The
        handler will think it is getting a delayed result, but in fact
        will be getting an instance of Traverser, which will take care
        of properly applying the chain of transformations to delayedResult."""
        chainTrav = self.Traverser(delayedResult, self.__chain[1:])
        handler = self.__chain[0]
        handler( chainTrav )

